// Matrix routines
// RjY/anARCHy

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "matrix.h"

// create a new matrix, initialised to zero
Matrix *matrix_new(Dimension N)
{
	int i,j;
	MElts elt;
	Matrix *mt;

	elt = calloc(N, sizeof(VElts));
	for (i = 0; i < N; i++) {
		elt[i] = calloc(N, sizeof(Real));
		for (j = 0; j < N; j++)
			elt[i][j] = 0.0; // 0.0 more like o_O
	}
	mt = (Matrix *)malloc(sizeof(Matrix));
	mt->N = N;
	mt->m = elt;
	return mt;
}

// free a matrix
void matrix_free(Matrix *M)
{
	int i;

	for (i = 0; i < M->N; i++)
		free(M->m[i]);
	free(M->m);
	free(M);
}

// create a new matrix, initialised from supplied MElts 
Matrix *matrix_new_elts(Dimension N, MElts elts)
{
	int i, j;
	Matrix *M = matrix_new(N); // initialised to zero

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
			M->m[i][j] = elts[i][j];
	return M;
}

// get element
Real matrix_get_elt(Matrix *M, int i, int j)
{
	if (!M || i>=M->N || j>=M->N)
		return 0;
	return M->m[i][j];
}

// set element to value
void matrix_set_elt(Matrix *M, int i, int j, Real v)
{
	if (M && i<M->N && j<M->N)
		M->m[i][j] = v;
}

// duplicate a matrix. nicely trivial isn't it
Matrix *matrix_dup(Matrix *M)
{
	return matrix_new_elts(M->N, M->m);
}

// dump the contents of a matrix to stdout (debugging routine)
void matrix_print(Matrix *M)
{
	int i, j;

	for (i = 0; i < M->N; i++) {
		printf("(");
		for (j = 0; j < M->N; j++)
			printf("%8.3f", M->m[i][j]); // FIXME
		printf(")\n");
	}
}

// Adds the second matrix onto the first
Matrix *matrix_add(Matrix *A, Matrix *B)
{
	int i, j;

	if (A->N != B->N)
		return NULL;

	for (i = 0; i < A->N; i++)
		for (j = 0; j < A->N; j++)
			A->m[i][j] += B->m[i][j];
	return A;
}

// Multiplies the matrix by a scalar
Matrix *matrix_scalarmult(Matrix *A, Real t)
{
	int i, j;

	for (i = 0; i < A->N; i++)
		for (j = 0; j < A->N; j++)
			A->m[i][j] *= t;
	return A;
}

// Returns the multiple of two NxN matrices
Matrix *matrix_mult(Matrix *A, Matrix *B)
{
	Matrix *M;
	int i, j, k;
	Real S;

	if (A->N != B->N)
		return NULL;

	M = matrix_new(A->N);

	for (i = 0; i < M->N; i++)
		for (j = 0; j < M->N; j++) {
			S = 0.0;
			for (k = 0; k < M->N; k++)
				S += A->m[i][k] * B->m[k][j];
			M->m[i][j] = S;
		}

	return M;
}

// transpose a matrix in place. Returns a pointer to M
Matrix *matrix_transpose(Matrix *M)
{
	int i, j;
	Real tmp;

	for (i = 1; i < M->N; i++)
		for (j = 0; j < i; j++) {
			tmp = M->m[i][j];
			M->m[i][j] = M->m[j][i];
			M->m[j][i] = tmp;
		}
	return M;
}

// Given an Nd matrix return an (N-1)d matrix removing the specified row/column
#define SKIP(n,skip) ( ((n)<(skip)) ? (n) : (n)+1)
Matrix *matrix_chomp(Matrix *M, int row, int col)
{
	Matrix *C;
	int i, j;

	if (M->N < 2)
		return NULL;
	C = matrix_new(M->N - 1);
	
	for (i = 0; i < C->N; i++)
		for (j = 0; j < C->N; j++)
			C->m[i][j] = M->m[SKIP(i,row)][SKIP(j,col)];
	return C;
}
#undef SKIP

// Recursive routine to calculate a matrix determinant.
Real matrix_det(Matrix *M)
{
	int i;
	Matrix *S;
	Real det = 0.0;

	if (M->N == 1)
		return M->m[0][0];

	for (i = 0; i < M->N; i++) {
		S = matrix_chomp(M, i, 0);
		det += (1-2*(i%2)) * M->m[i][0] * matrix_det(S);
		matrix_free(S);
	}
	return det;
}

// Find the cofactor matrix of a given matrix (this is expensive)
Matrix *matrix_cofactor(Matrix *M)
{
	Matrix *C, *c;
	int i, j;

	C = matrix_new(M->N);
	if (C->N == 1)
		C->m[0][0] = 1.0;
	else
		for (i = 0; i < M->N; i++)
			for (j = 0; j < M->N; j++) {
				c = matrix_chomp(M, i, j);
				C->m[i][j] = (1-2*((i+j)%2)) * matrix_det(c);
				matrix_free(c);
			}
	return C;
}

// Find the inverse of a given matrix
Matrix *matrix_invert(Matrix *M) {
	Matrix *I;
	Real det;

	// calculate determinant of M
	det = matrix_det(M);

	if (det == 0) // FIXME do this better.
		return NULL; // not invertible

	// get cofactor matrix
	I = matrix_cofactor(M);

	// transpose it to get adjoint matrix
	I = matrix_transpose(I);

	// multiply it by the reciprocal of the determinant
	I = matrix_scalarmult(I, 1.0/det);

	// HEY GUYS WHATS GOIN ON IN THIS INVERTED MATRIX
	return I;
}
